@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2002, 2003, 2004, 2005, 2006, 2007, 2009, 2010 by        */
+/*  Copyright 1996-2007, 2009-2015 by                                       */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -10,11 +10,9 @@
 /*                                                                          */
 /****************************************************************************/
 
-#include <ft2build.h>
-#include FT_FREETYPE_H
 
-#include "common.h"
 #include "ftcommon.h"
+#include "common.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -22,19 +20,26 @@
 #include <stdarg.h>
 #include <math.h>
 
+#include FT_LCD_FILTER_H
+
 #define CELLSTRING_HEIGHT  8
-#define MAXPTSIZE  500                 /* dtp */
+#define MAXPTSIZE          500   /* dtp */
 
 
   static char*  Text = (char *)"The quick brown fox jumps over the lazy dog";
 
-  enum {
+  enum
+  {
     RENDER_MODE_STRING,
     RENDER_MODE_KERNCMP,
     N_RENDER_MODES
   };
 
-  static struct {
+  static struct  status_
+  {
+    int          width;
+    int          height;
+
     int          render_mode;
     FT_Encoding  encoding;
     int          res;
@@ -44,40 +49,59 @@
 
     FTDemo_String_Context  sc;
 
-    FT_Byte      gamma_ramp[256];
+    FT_Byte      gamma_ramp[256];   /* for show only */
     FT_Matrix    trans_matrix;
     int          font_index;
     char*        header;
     char         header_buffer[256];
 
-  } status = { RENDER_MODE_STRING, FT_ENCODING_UNICODE, 72, 48, 2.0, 0 };
+  } status = { DIM_X, DIM_Y,
+               RENDER_MODE_STRING, FT_ENCODING_UNICODE, 72, 48, GAMMA, 0,
+               { 0, 0, 0, 0, NULL },
+               { 0 }, { 0, 0, 0, 0 }, 0, NULL, { 0 } };
 
   static FTDemo_Display*  display;
   static FTDemo_Handle*   handle;
 
 
-/****************************************************************************/
-/****************************************************************************/
-/****************************************************************************/
-/****                                                                    ****/
-/****                     E V E N T   H A N D L I N G                    ****/
-/****                                                                    ****/
-/****************************************************************************/
-/****************************************************************************/
-/****************************************************************************/
+  /*************************************************************************/
+  /*************************************************************************/
+  /*************************************************************************/
+  /****                                                                 ****/
+  /****                   E V E N T   H A N D L I N G                   ****/
+  /****                                                                 ****/
+  /*************************************************************************/
+  /*************************************************************************/
+  /*************************************************************************/
 
   static void
   event_help( void )
   {
+    char  buf[256];
+    char  version[64];
+
+    const char*  format;
+    FT_Int       major, minor, patch;
+
     grEvent  dummy_event;
 
 
+    FT_Library_Version( handle->library, &major, &minor, &patch );
+
+    format = patch ? "%d.%d.%d" : "%d.%d";
+    sprintf( version, format, major, minor, patch );
+
     FTDemo_Display_Clear( display );
+    grSetLineHeight( 10 );
     grGotoxy( 0, 0 );
     grSetMargin( 2, 1 );
     grGotobitmap( display->bitmap );
 
-    grWriteln( "FreeType String Viewer - part of the FreeType test suite" );
+    sprintf( buf,
+             "FreeType String Viewer - part of the FreeType %s test suite",
+             version );
+
+    grWriteln( buf );
     grLn();
     grWriteln( "This program is used to display a string of text using" );
     grWriteln( "the new convenience API of the FreeType 2 library." );
@@ -90,14 +114,14 @@
     grWriteln( "  b         : toggle embedded bitmaps (and disable rotation)" );
     grWriteln( "  f         : toggle forced auto-hinting" );
     grWriteln( "  h         : toggle outline hinting" );
-    grWriteln( "  l         : toggle low precision rendering" );
     grLn();
     grWriteln( "  1-2       : select rendering mode" );
+    grWriteln( "  l         : cycle through LCD modes" );
     grWriteln( "  k         : cycle through kerning modes" );
     grWriteln( "  t         : cycle through kerning degrees" );
+    grWriteln( "  space     : cycle through color" );
     grWriteln( "  V         : toggle vertical rendering" );
     grLn();
-    grWriteln( "  G         : toggle gamma correction" );
     grWriteln( "  g         : increase gamma by 0.1" );
     grWriteln( "  v         : decrease gamma by 0.1" );
     grLn();
@@ -114,7 +138,6 @@
     grWriteln( "  F7        : big rotate counter-clockwise" );
     grWriteln( "  F8        : big rotate clockwise" );
     grLn();
-    grLn();
     grWriteln( "press any key to exit this help screen" );
 
     grRefreshSurface( display->surface );
@@ -126,7 +149,7 @@
   event_font_change( int  delta )
   {
     if ( status.font_index + delta >= handle->num_fonts ||
-         status.font_index + delta < 0 )
+         status.font_index + delta < 0                  )
       return;
 
     status.font_index += delta;
@@ -147,7 +170,12 @@
     FT_Fixed  sinus;
 
 
-    status.angle = ( status.angle + delta ) % 360;
+    status.angle += delta;
+
+    if ( status.angle <= -180 )
+      status.angle += 360;
+    if ( status.angle > 180 )
+      status.angle -= 360;
 
     if ( status.angle == 0 )
     {
@@ -158,10 +186,7 @@
 
     status.sc.matrix = &status.trans_matrix;
 
-    if ( status.angle < 0 )
-      status.angle += 360;
-
-    radian  = status.angle * 3.14159 / 180.0;
+    radian  = status.angle * 3.14159265 / 180.0;
     cosinus = (FT_Fixed)( cos( radian ) * 65536.0 );
     sinus   = (FT_Fixed)( sin( radian ) * 65536.0 );
 
@@ -173,10 +198,62 @@
 
 
   static void
-  event_gamma_change( double delta )
+  event_lcdmode_change( void )
   {
-    int i;
-    double gamma_inv;
+    const char  *lcd_mode;
+
+
+    handle->lcd_mode++;
+
+    switch ( handle->lcd_mode )
+    {
+    case LCD_MODE_LIGHT:
+      lcd_mode = " light AA";
+      break;
+    case LCD_MODE_RGB:
+      lcd_mode = " LCD (horiz. RGB)";
+      break;
+    case LCD_MODE_BGR:
+      lcd_mode = " LCD (horiz. BGR)";
+      break;
+    case LCD_MODE_VRGB:
+      lcd_mode = " LCD (vert. RGB)";
+      break;
+    case LCD_MODE_VBGR:
+      lcd_mode = " LCD (vert. BGR)";
+      break;
+    default:
+      handle->lcd_mode = 0;
+      lcd_mode = " normal AA";
+    }
+
+    sprintf( status.header_buffer, "mode changed to %s", lcd_mode );
+    status.header = status.header_buffer;
+  }
+
+
+  static void
+  event_color_change( void )
+  {
+    static int     i = 0;
+    unsigned char  r = i & 4 ? 0xff : 0;
+    unsigned char  g = i & 2 ? 0xff : 0;
+    unsigned char  b = i & 1 ? 0xff : 0;
+
+
+    display->back_color = grFindColor( display->bitmap,  r,  g,  b, 0xff );
+    display->fore_color = grFindColor( display->bitmap, ~r, ~g, ~b, 0xff );
+
+    i++;
+  }
+
+
+  static void
+  event_gamma_change( double  delta )
+  {
+    int     i;
+    double  gamma_inv;
+
 
     status.gamma += delta;
 
@@ -185,33 +262,32 @@
     else if ( status.gamma < 0.1 )
       status.gamma = 0.1;
 
-    sprintf( status.header_buffer, "gamma changed to %.1f", status.gamma );
-    status.header = status.header_buffer;
+    grSetGlyphGamma( status.gamma );
 
-    gamma_inv = 1.0f / status.gamma;
+    gamma_inv = 1. / status.gamma;
 
     for ( i = 0; i < 256; i++ )
-      status.gamma_ramp[i] = (FT_Byte)( pow( (double)i / 255.0f, gamma_inv )
-                                        * 255.0f );
+      status.gamma_ramp[i] = (FT_Byte)( pow( (double)i / 255., gamma_inv )
+                                        * 255. + 0.5 );
   }
 
 
   static void
-  event_size_change( int delta )
+  event_size_change( int  delta )
   {
     status.ptsize += delta;
 
-    if ( status.ptsize < 1*64 )
-      status.ptsize = 1*64;
-    else if ( status.ptsize > MAXPTSIZE*64 )
-      status.ptsize = MAXPTSIZE*64;
+    if ( status.ptsize < 1 * 64 )
+      status.ptsize = 1 * 64;
+    else if ( status.ptsize > MAXPTSIZE * 64 )
+      status.ptsize = MAXPTSIZE * 64;
 
     FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
   }
 
 
   static void
-  event_render_mode_change( int delta )
+  event_render_mode_change( int  delta )
   {
     if ( delta )
     {
@@ -237,7 +313,7 @@
   static int
   Process_Event( grEvent*  event )
   {
-    FTDemo_String_Context*  sc = &status.sc;
+    FTDemo_String_Context*  sc  = &status.sc;
     int                     ret = 0;
 
 
@@ -298,10 +374,7 @@
       break;
 
     case grKEY( 'l' ):
-      handle->low_prec = !handle->low_prec;
-      status.header    = handle->low_prec
-                         ? (char *)"rendering precision is now forced to low"
-                         : (char *)"rendering precision is now normal";
+      event_lcdmode_change();
 
       FTDemo_Update_Current_Flags( handle );
       break;
@@ -328,6 +401,10 @@
             : (char *)"tight track kerning active";
       break;
 
+    case grKeySpace:
+      event_color_change();
+      break;
+
     case grKEY( 'V' ):
       sc->vertical  = !sc->vertical;
       status.header = sc->vertical
@@ -335,13 +412,6 @@
                       : (char *)"using horizontal layout";
       break;
 
-    case grKEY( 'G' ):
-      sc->gamma_ramp = sc->gamma_ramp ? NULL : status.gamma_ramp;
-      status.header  = sc->gamma_ramp
-                       ? (char *)"gamma correction is now on"
-                       : (char *)"gamma correction is now off";
-      break;
-
     case grKEY( 'g' ):
       event_gamma_change( 0.1 );
       break;
@@ -360,7 +430,7 @@
 
     case grKeyUp:       event_size_change(   64 ); break;
     case grKeyDown:     event_size_change(  -64 ); break;
-    case grKeyPageUp:   event_size_change(  640); break;
+    case grKeyPageUp:   event_size_change(  640 ); break;
     case grKeyPageDown: event_size_change( -640 ); break;
 
     case grKeyLeft:  event_angle_change(    -3 ); break;
@@ -381,8 +451,10 @@
                    grBitmap*  bitmap )
   {
     int       i, x, y;
+    int       bpp = bitmap->pitch / bitmap->width;
     FT_Byte*  p = (FT_Byte*)bitmap->buffer;
 
+
     if ( bitmap->pitch < 0 )
       p += -bitmap->pitch * ( bitmap->rows - 1 );
 
@@ -390,7 +462,7 @@
     y = ( bitmap->rows + 256 ) / 2;
 
     for (i = 0; i < 256; i++)
-      p[bitmap->pitch * ( y - gamma_ramp[i] ) + ( x + i )] = 80;
+      p[bitmap->pitch * ( y - gamma_ramp[i] ) + bpp * ( x + i )] ^= 0xFF;
   }
 
 
@@ -417,14 +489,17 @@
                  "%.50s %.50s (file `%.100s')", face->family_name,
                  face->style_name, basename );
         break;
+
       case FT_Err_Invalid_Pixel_Size:
         sprintf( status.header_buffer, "Invalid pixel size (file `%.100s')",
                  basename );
         break;
+
       case FT_Err_Invalid_PPem:
         sprintf( status.header_buffer, "Invalid ppem value (file `%.100s')",
                  basename );
         break;
+
       default:
         sprintf( status.header_buffer, "File `%.100s': error 0x%04x", basename,
             (FT_UShort)error_code );
@@ -437,8 +512,8 @@
     grWriteCellString( display->bitmap, 0, 0,
                        status.header, display->fore_color );
 
-    sprintf( status.header_buffer, "at %g points, angle = %d",
-             status.ptsize/64.0, status.angle );
+    sprintf( status.header_buffer, "at %g points, angle = %d, gamma = %g",
+             status.ptsize / 64.0, status.angle, status.gamma );
     grWriteCellString( display->bitmap, 0, CELLSTRING_HEIGHT,
                        status.header_buffer, display->fore_color );
 
@@ -449,17 +524,39 @@
   static void
   usage( char*  execname )
   {
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "ftstring: string viewer -- part of the FreeType project\n" );
-    fprintf( stderr,  "-------------------------------------------------------\n" );
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "Usage: %s [options below] ppem fontname[.ttf|.ttc] ...\n",
+    fprintf( stderr,
+      "\n"
+      "ftstring: string viewer -- part of the FreeType project\n"
+      "-------------------------------------------------------\n"
+      "\n" );
+    fprintf( stderr,
+      "Usage: %s [options] pt font ...\n"
+      "\n",
              execname );
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "  -e enc      specify encoding tag (default: unic)\n" );
-    fprintf( stderr,  "  -r R        use resolution R dpi (default: 72 dpi)\n" );
-    fprintf( stderr,  "  -m message  message to display\n" );
-    fprintf( stderr,  "\n" );
+    fprintf( stderr,
+      "  pt        The point size for the given resolution.\n"
+      "            If resolution is 72dpi, this directly gives the\n"
+      "            ppem value (pixels per EM).\n" );
+    fprintf( stderr,
+      "  font      The font file(s) to display.\n"
+      "            For Type 1 font files, ftstring also tries to attach\n"
+      "            the corresponding metrics file (with extension\n"
+      "            `.afm' or `.pfm').\n"
+      "\n" );
+    fprintf( stderr,
+      "  -w W      Set the window width to W pixels (default: %dpx).\n"
+      "  -h H      Set the window height to H pixels (default: %dpx).\n"
+      "\n",
+             DIM_X, DIM_Y );
+    fprintf( stderr,
+      "  -r R      Use resolution R dpi (default: 72dpi).\n"
+      "  -e enc    Specify encoding tag (default: no encoding).\n"
+      "            Common values: `unic' (Unicode), `symb' (symbol),\n"
+      "            `ADOB' (Adobe standard), `ADBC' (Adobe custom).\n"
+      "  -m text   Use `text' for rendering.\n"
+      "\n"
+      "  -v        Show version.\n"
+      "\n" );
 
     exit( 1 );
   }
@@ -477,7 +574,7 @@
 
     while ( 1 )
     {
-      option = getopt( *argc, *argv, "e:m:r:" );
+      option = getopt( *argc, *argv, "e:h:m:r:vw:" );
 
       if ( option == -1 )
         break;
@@ -488,9 +585,9 @@
         status.encoding = FTDemo_Make_Encoding_Tag( optarg );
         break;
 
-      case 'r':
-        status.res = atoi( optarg );
-        if ( status.res < 1 )
+      case 'h':
+        status.height = atoi( optarg );
+        if ( status.height < 1 )
           usage( execname );
         break;
 
@@ -500,6 +597,33 @@
         Text = optarg;
         break;
 
+      case 'r':
+        status.res = atoi( optarg );
+        if ( status.res < 1 )
+          usage( execname );
+        break;
+
+      case 'v':
+        {
+          FT_Int  major, minor, patch;
+
+
+          FT_Library_Version( handle->library, &major, &minor, &patch );
+
+          printf( "ftstring (FreeType) %d.%d", major, minor );
+          if ( patch )
+            printf( ".%d", patch );
+          printf( "\n" );
+          exit( 0 );
+        }
+        /* break; */
+
+      case 'w':
+        status.width = atoi( optarg );
+        if ( status.width < 1 )
+          usage( execname );
+        break;
+
       default:
         usage( execname );
         break;
@@ -512,7 +636,7 @@
     if ( *argc <= 1 )
       usage( execname );
 
-    status.ptsize = (int)(atof( *argv[0] ) * 64.0);
+    status.ptsize = (int)( atof( *argv[0] ) * 64.0 );
     if ( status.ptsize == 0 )
       status.ptsize = 64;
 
@@ -528,17 +652,20 @@
     grEvent  event;
 
 
+    /* Initialize engine */
+    handle = FTDemo_New();
+
     parse_cmdline( &argc, &argv );
 
-    /* Initialize engine */
-    handle = FTDemo_New( status.encoding );
+    FT_Library_SetLcdFilter( handle->library, FT_LCD_FILTER_LIGHT );
 
+    handle->encoding  = status.encoding;
     handle->use_sbits = 0;
     FTDemo_Update_Current_Flags( handle );
 
     for ( ; argc > 0; argc--, argv++ )
     {
-      error = FTDemo_Install_Font( handle, argv[0] );
+      error = FTDemo_Install_Font( handle, argv[0], 0 );
 
       if ( error )
       {
@@ -553,24 +680,26 @@
     if ( handle->num_fonts == 0 )
       PanicZ( "could not open any font file" );
 
-    display = FTDemo_Display_New( gr_pixel_mode_gray );
-    display->back_color.value = 0;
-    display->fore_color.value = 0xff;
+    display = FTDemo_Display_New( gr_pixel_mode_rgb24,
+                                  status.width, status.height );
 
     if ( !display )
       PanicZ( "could not allocate display surface" );
 
     grSetTitle( display->surface,
-                "FreeType String Viewer - press F1 for help" );
+                "FreeType String Viewer - press ? for help" );
 
+    event_color_change();
     event_gamma_change( 0 );
     event_font_change( 0 );
     status.header = 0;
 
-    for ( ;; )
+    do
     {
       FTDemo_Display_Clear( display );
 
+      gamma_ramp_draw( status.gamma_ramp, display->bitmap );
+
       switch ( status.render_mode )
       {
       case RENDER_MODE_STRING:
@@ -585,13 +714,12 @@
         {
           FTDemo_String_Context  sc = status.sc;
           FT_Int                 x, y;
-          FT_UInt                height;
+          FT_Int                 height;
 
 
           x = 55;
 
-          /* whatever.. */
-          height = status.ptsize * status.res / 72;
+          height = ( status.ptsize * status.res / 72 + 32 ) >> 6;
           if ( height < CELLSTRING_HEIGHT )
             height = CELLSTRING_HEIGHT;
 
@@ -608,7 +736,6 @@
                              "none", display->fore_color );
           error = FTDemo_String_Draw( handle, display, &sc, x, y );
 
-
           /* Second line: track kern only */
           sc.kerning_degree = status.sc.kerning_degree;
 
@@ -618,9 +745,8 @@
                              "track", display->fore_color );
           error = FTDemo_String_Draw( handle, display, &sc, x, y );
 
-
           /* Third line: track kern + pair kern */
-          sc.kerning_mode      = status.sc.kerning_mode;
+          sc.kerning_mode = status.sc.kerning_mode;
 
           y += height;
           grWriteCellString( display->bitmap, 5,
@@ -631,16 +757,11 @@
         break;
       }
 
-      if ( !error && status.sc.gamma_ramp )
-        gamma_ramp_draw( status.gamma_ramp, display->bitmap );
-
       write_header( error );
 
       status.header = 0;
       grListenSurface( display->surface, 0, &event );
-      if ( Process_Event( &event ) )
-        break;
-    }
+    } while ( !Process_Event( &event ) );
 
     printf( "Execution completed successfully.\n" );
 
@@ -648,7 +769,7 @@
     FTDemo_Done( handle );
     exit( 0 );      /* for safety reasons */
 
-    return 0;       /* never reached */
+    /* return 0; */ /* never reached */
   }
 
 
